@@ -121,6 +121,7 @@ copy (HTMLObject *self,
 	HTML_CLUEFLOW (dest)->clear = HTML_CLUEFLOW (self)->clear;
 	HTML_CLUEFLOW (dest)->item_color = HTML_CLUEFLOW (self)->item_color;
 	HTML_CLUEFLOW (dest)->indent_width = HTML_CLUEFLOW (self)->indent_width;
+	HTML_CLUEFLOW (dest)->dir = HTML_CLUEFLOW (self)->dir;
 	
 	if (HTML_CLUEFLOW (dest)->item_color)
 		html_color_ref (HTML_CLUEFLOW (dest)->item_color);
@@ -226,6 +227,32 @@ get_recursive_length (HTMLObject *self)
 	return (*HTML_OBJECT_CLASS (parent_class)->get_recursive_length) (self) + (self->next ? 1 : 0);
 }
 
+static gboolean
+prev_flow_in_cluevs (HTMLObject *o)
+{
+	if (o->prev)
+		return TRUE;
+
+	o = o->parent;
+
+	/* go up the tree to find cluev with prev */
+	while (o && HTML_IS_CLUEV (o) && !o->prev)
+		o = o->parent;
+
+	/* go down the tree to find last clueflow */
+	if (o && o->prev) {
+		o = o->prev;
+
+		while (o && HTML_IS_CLUEV (o))
+			o = HTML_CLUE (o)->tail;
+
+		if (o && HTML_IS_CLUEFLOW (o))
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
 static HTMLObject *
 op_helper (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GList *right, guint *len, gboolean cut)
 {
@@ -237,7 +264,7 @@ op_helper (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left,
 		return NULL; 
 	if (!from && (*len || !(self->prev && HTML_IS_CLUEFLOW (self->prev) && HTML_IS_TABLE (HTML_CLUE (self->prev)->tail))))
 	(*len) ++; */
-	if (!from && self->prev) {
+	if (!from && prev_flow_in_cluevs (self)) {
 		(*len) ++;
 		/* if (cut)
 		   e->cursor->position --; */
@@ -348,6 +375,7 @@ merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **
 		HTML_CLUE  (cf1)->halign = HTML_CLUE (cf2)->halign;
 		HTML_CLUE  (cf1)->valign = HTML_CLUE (cf2)->valign;
 		html_object_copy_data_from_object (self, with);
+		cf1->dir = cf2->dir;
 	}
 
 	rv = (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right, cursor);
@@ -558,13 +586,15 @@ get_level_indent (HTMLClueFlow *flow,
 		  gint level,
 		  HTMLPainter *painter)
 {
+	HTMLDirection dir = html_object_get_direction (HTML_OBJECT (flow));
 	guint indent = 0;
 	gint i = 0;
 
 	if (flow->levels->len > 0 || ! is_item (flow)) {
 		guint cite_width, indent_width;
 
-		cite_width   = html_painter_get_block_cite_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
+		cite_width = html_painter_get_block_cite_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL, dir)
+			+ html_painter_get_space_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
 		indent_width = html_painter_get_block_indent_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
 		
 		while (i <= level) {
@@ -658,18 +688,37 @@ calc_min_width (HTMLObject *o,
 }
 
 static gint
-pref_right_margin (HTMLPainter *p, HTMLClueFlow *clueflow, HTMLObject *o, gint y, gboolean with_aligned) 
+pref_left_margin (HTMLPainter *p, HTMLObject *o, int indent)
 {
-	gint fixed_margin = html_object_get_right_margin (o, p, y, with_aligned);
+	int margin = html_object_get_left_margin (o->parent, p, o->y, TRUE);
 
-	/* FIXME: this hack lets us wrap the display at 72 characters when we are using
-	   a plain painter */
-	  
-	if (clueflow->style == HTML_CLUEFLOW_STYLE_PRE || ! HTML_IS_PLAIN_PAINTER(p))
-		return fixed_margin;
+	if (html_object_get_direction (o) == HTML_DIRECTION_RTL) {
+		if (HTML_CLUEFLOW (o)->style != HTML_CLUEFLOW_STYLE_PRE && HTML_IS_PLAIN_PAINTER(p))
+			return MAX (margin, o->width - (int)(72 * (MAX (html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL),
+									html_painter_get_e_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL)))));
+	} else {
+		if (indent > margin)
+			margin = indent;
+	}
 
-	return MIN (fixed_margin, 72 * (MAX (html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL),
-					     html_painter_get_e_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL))));
+	return margin;
+}
+
+static gint
+pref_right_margin (HTMLPainter *p, HTMLObject *o, int indent)
+{
+	gint margin = html_object_get_right_margin (o->parent, p, o->y, TRUE);
+
+	if (html_object_get_direction (o) != HTML_DIRECTION_RTL) {
+		if (HTML_CLUEFLOW (o)->style != HTML_CLUEFLOW_STYLE_PRE && HTML_IS_PLAIN_PAINTER(p))
+			return MIN (margin, 72 * (MAX (html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL),
+						       html_painter_get_e_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL))));
+	} else {
+		if (indent > o->width - margin)
+			margin = o->width - indent;
+	}
+
+	return margin;
 }
 
 static void
@@ -693,10 +742,8 @@ add_clear_area (GList **changed_objs, HTMLObject *o, gint x, gint w)
 static void
 calc_margins (HTMLObject *o, HTMLPainter *painter, gint indent, gint *lmargin, gint *rmargin)
 {
-	*lmargin = html_object_get_left_margin (o->parent, painter, o->y, TRUE);
-	if (indent > *lmargin)
-		*lmargin = indent;
-	*rmargin = pref_right_margin (painter, HTML_CLUEFLOW (o), o->parent, o->y, TRUE);
+	*lmargin = pref_left_margin (painter, o, indent);
+	*rmargin = pref_right_margin (painter, o, indent);
 }
 
 static inline gint
@@ -1123,25 +1170,25 @@ get_item_marker_str (HTMLClueFlow *flow, gboolean ascii_only)
 }
 
 static void
-draw_gt_line (HTMLObject *cur, HTMLPainter *p, gint offset, gint x, gint y)
+draw_cite_line (HTMLObject *cur, HTMLPainter *p, char *cite_str, gint offset, gint x, gint y)
 {
 	gint cy, w, a, d;
 
 	/* FIXME: cache items and glyphs? */
-	html_painter_calc_text_size (p, HTML_BLOCK_CITE, strlen (HTML_BLOCK_CITE),
+	html_painter_calc_text_size (p, cite_str, strlen (cite_str),
 				     &w, &a, &d);
 
 	cy = offset;
 	while (cy + a <= cur->ascent) {
 		/* FIXME: cache items and glyphs? */
-		html_painter_draw_text (p, x, y + cur->y - cy, HTML_BLOCK_CITE, 1);
+		html_painter_draw_text (p, x, y + cur->y - cy, cite_str, 1);
 		cy += a + d;
 	}
 
 	cy = - offset + a + d;
 	while (cy + d <= cur->descent) {
 		/* FIXME: cache items and glyphs? */
-		html_painter_draw_text (p, x, y + cur->y + cy, HTML_BLOCK_CITE, 1);
+		html_painter_draw_text (p, x, y + cur->y + cy, cite_str, 1);
 		cy += a + d;
 	}
 }
@@ -1158,10 +1205,11 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 	int last_indent = 0;
 	gint pixel_size = html_painter_get_pixel_size (painter);
 	gboolean is_plain = HTML_IS_PLAIN_PAINTER (painter);
+	HTMLDirection dir = html_object_get_direction (self);
 	HTMLEngine *e;
 
 	if (painter->widget && GTK_IS_HTML (painter->widget))
-		e = GTK_HTML (painter->widget)->engine;
+		e = html_object_engine (self, GTK_HTML (painter->widget)->engine);
 	else
 		return;
 	
@@ -1174,7 +1222,10 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 										   painter, HTMLLinkColor)->color);
 		if (is_cite (flow, i)) {
 			if (!is_plain) {
-				area.x = self->x + indent - 5 * pixel_size;
+				if (dir == HTML_DIRECTION_RTL)
+					area.x = self->x + self->width - indent + 3 * pixel_size;
+				else
+					area.x = self->x + indent - 5 * pixel_size;
 				area.width = 2 * pixel_size;
 				area.y = self->y - self->ascent;
 				area.height = self->ascent + self->descent;
@@ -1192,7 +1243,9 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 							paint.width, paint.height);
 			} else {
 				HTMLObject *cur = HTML_CLUE (self)->head;
-				gint baseline = 0;
+				int x_pos, baseline = 0;
+				char *cite_str = dir == HTML_DIRECTION_RTL ? HTML_BLOCK_CITE_RTL : HTML_BLOCK_CITE_LTR;
+
 				while (cur) {
 					if (cur->y != 0) {
 						baseline = cur->y;
@@ -1203,12 +1256,16 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 
 
 				/* draw "> " quote characters in the plain case */ 
-				html_painter_set_font_style (painter, 
-							     html_clueflow_get_default_font_style (flow));
-						
+				html_painter_set_font_style (painter, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED);
 				html_painter_set_font_face  (painter, NULL);
-				draw_gt_line (self, painter, self->ascent - baseline,
-					      self->x + tx + last_indent, ty);
+
+				if (dir == HTML_DIRECTION_RTL)
+					x_pos = self->x + self->width - last_indent;
+				else
+					x_pos = self->x + last_indent;
+
+				draw_cite_line (self, painter, cite_str, self->ascent - baseline,
+						x_pos + tx, ty);
 			}
 		}
 	}
@@ -1223,7 +1280,7 @@ draw_item (HTMLObject *self, HTMLPainter *painter, gint x, gint y, gint width, g
 	HTMLEngine *e;
 
 	if (painter->widget && GTK_IS_HTML (painter->widget))
-		e = GTK_HTML (painter->widget)->engine;
+		e = html_object_engine (self, GTK_HTML (painter->widget)->engine);
 	else
 		return;
 
@@ -1242,16 +1299,20 @@ draw_item (HTMLObject *self, HTMLPainter *painter, gint x, gint y, gint width, g
 
 	marker = get_item_marker_str (flow, HTML_IS_PLAIN_PAINTER (painter));
 	if (marker) {
-		gint width, len, asc, dsc;
+		gint width, len, asc, dsc, x_pos, space_width;
 		
 		len   = g_utf8_strlen (marker, -1);
 		html_painter_set_font_style (painter, html_clueflow_get_default_font_style (flow));
 		html_painter_set_font_face  (painter, NULL);
 		/* FIXME: cache items and glyphs? */
 		html_painter_calc_text_size (painter, marker, len, &width, &asc, &dsc);
-		width += html_painter_get_space_width (painter, html_clueflow_get_default_font_style (flow), NULL);
+		space_width = html_painter_get_space_width (painter, html_clueflow_get_default_font_style (flow), NULL);
 		/* FIXME: cache items and glyphs? */
-		html_painter_draw_text (painter, self->x + first->x - width + tx,
+		if (html_object_get_direction (self) == HTML_DIRECTION_RTL)
+			x_pos = self->x + first->x + first->width + space_width;
+		else
+			x_pos = self->x + first->x - width - space_width;
+		html_painter_draw_text (painter, x_pos + tx,
 					self->y - self->ascent + first->y + ty,
 					marker, len);
 	}
@@ -1545,6 +1606,28 @@ save_indent_string (HTMLClueFlow *self, HTMLEngineSaveState *state, const char *
 	return retval;
 }
 
+static char *
+get_p_str (HTMLClueFlow *self, HTMLEngineSaveState *state)
+{
+	char *p_str = NULL;
+
+	if (self->dir != html_object_get_direction (state->engine->clue)) {
+		switch (self->dir) {
+		case HTML_DIRECTION_RTL:
+			p_str = "<P DIR=RTL>\n";
+			break;
+		case HTML_DIRECTION_LTR:
+			if (html_object_get_direction (state->engine->clue) != HTML_DIRECTION_DERIVED)
+				p_str = "<P DIR=LTR>\n";
+			break;
+		default:
+			;
+		}
+	}
+
+	return p_str;
+}
+
 static gboolean
 write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state) 
 {
@@ -1552,13 +1635,37 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 	HTMLClueFlow *next = NULL;
 	HTMLClueFlow *prev = NULL;
 	HTMLHAlignType halign;
-	
+	char *br_str = "<BR>\n";
+
 	if (HTML_IS_CLUEFLOW (HTML_OBJECT (self)->next))
 		next = HTML_CLUEFLOW (HTML_OBJECT (self)->next);
 	    
 	if (HTML_IS_CLUEFLOW (HTML_OBJECT (self)->prev))
 		prev = HTML_CLUEFLOW (HTML_OBJECT (self)->prev);
 
+	if (next && next->dir != html_object_get_direction (state->engine->clue)) {
+		switch (next->dir) {
+		case HTML_DIRECTION_RTL:
+			br_str = "<BR DIR=RTL>\n";
+			break;
+		case HTML_DIRECTION_LTR:
+			if (html_object_get_direction (state->engine->clue) != HTML_DIRECTION_DERIVED)
+				br_str = "<BR DIR=LTR>\n";
+			break;
+		default:
+			;
+		}
+	}
+
+	if (!prev) {
+		char *p_str = get_p_str (self, state);
+
+		if (p_str) {
+			if (! html_engine_save_output_string (state, p_str))
+				return FALSE;
+		}
+	}
+
 	d = get_similar_depth (self, prev);
 	if (is_item (self)) {
 		char *li = get_list_start_tag (self);
@@ -1606,13 +1713,13 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 
 	if (is_item (self)) {
 		if (next && is_levels_equal (self, next) && !is_item (next) && !html_clueflow_contains_table (self)) {
-			if (!html_engine_save_output_string (state, "<BR>\n"))
+			if (!html_engine_save_output_string (state, br_str))
 				return FALSE;
 		} else if (!html_engine_save_output_string (state, "\n"))
 			return FALSE;
 	} else if (is_levels_equal (self, next) && self->style == next->style) {
 		if (self->style != HTML_CLUEFLOW_STYLE_PRE && !html_clueflow_contains_table (self)) {
-			if (!html_engine_save_output_string (state, "<BR>\n"))
+			if (!html_engine_save_output_string (state, br_str))
 				return FALSE;
 		} else {
 			if (!html_engine_save_output_string (state, "\n"))
@@ -1623,7 +1730,7 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 
 		if (self->style != HTML_CLUEFLOW_STYLE_PRE) {
 			if ((!html_clueflow_contains_table (self) && !end && next && self->style == next->style) || html_clueflow_is_empty (self)) {
-				if (!html_engine_save_output_string (state, "<BR>\n"))
+				if (!html_engine_save_output_string (state, br_str))
 					return FALSE;
 			} else {
 				if (!html_engine_save_output_string (state, "\n"))
@@ -1640,6 +1747,21 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 		}
 	}
 	
+	if (!next) {
+		if (HTML_OBJECT (self)->parent && html_object_is_clue (HTML_OBJECT (self)->parent)) {
+			HTMLObject *head = HTML_CLUE (HTML_OBJECT (self)->parent)->head;
+
+			if (head && HTML_IS_CLUEFLOW (head)) {
+				char *head_p_str = get_p_str (HTML_CLUEFLOW (head), state);
+
+				if (head_p_str) {
+					if (! html_engine_save_output_string (state, "</P>\n"))
+						return FALSE;
+				}
+			}
+		}
+	}
+
 	return TRUE;
 }
 
@@ -1731,7 +1853,7 @@ plain_padding (HTMLClueFlow *flow, GString *out, gboolean firstline)
 		for (i = 0; i < flow->levels->len; i++) {
 			switch (flow->levels->data[i]) {
 			case HTML_LIST_TYPE_BLOCKQUOTE_CITE:
-				APPEND_PLAIN (HTML_BLOCK_CITE);
+				APPEND_PLAIN (html_object_get_direction (HTML_OBJECT (flow)) == HTML_DIRECTION_RTL ? (" " HTML_BLOCK_CITE_RTL) : (HTML_BLOCK_CITE_LTR " "));
 				break;
 			case HTML_LIST_TYPE_GLOSSARY_DL:
 				break;
@@ -1816,6 +1938,7 @@ save_plain (HTMLObject *self,
 {
 	HTMLClueFlow *flow;
 	HTMLEngineSaveState *buffer_state;
+	HTMLDirection dir = html_object_get_direction (self);
 	GString *out = g_string_new ("");
 	gint pad;
 	gint align_pad;
@@ -1945,13 +2068,19 @@ save_plain (HTMLObject *self,
 
 				switch (html_clueflow_get_halignment (flow)) {
 				case HTML_HALIGN_RIGHT:
-					align_pad = max_width - width;
+					if (dir != HTML_DIRECTION_RTL)
+						align_pad = max_width - width;
+					else
+						align_pad = 0;
 					break;
 				case HTML_HALIGN_CENTER:
 					align_pad = (max_width - width) / 2;
 					break;
 				default:
-					align_pad = 0;
+					if (dir != HTML_DIRECTION_RTL)
+						align_pad = 0;
+					else
+						align_pad = max_width - width;
 					break;
 				}
 			
@@ -1982,7 +2111,7 @@ save_plain (HTMLObject *self,
 			g_string_append (out, "\n");
 		}
 	}
-	html_engine_save_buffer_free (buffer_state);
+	html_engine_save_buffer_free (buffer_state, TRUE);
 		
 	if (!html_engine_save_output_string (state, "%s", out->str)) {
 		g_string_free (out, TRUE);
@@ -2316,17 +2445,10 @@ html_clueflow_real_get_direction (HTMLObject *o)
 		HTMLObject *child;
 		for (child = HTML_CLUE (o)->head; child; child = child->next) {
 			if (HTML_IS_TEXT (child)) {
-				PangoDirection pdir = html_text_get_pango_direction (HTML_TEXT (child));
-				switch (pdir) {
-				case PANGO_DIRECTION_RTL:
-					dir = HTML_DIRECTION_RTL;
-					break;
-				case PANGO_DIRECTION_LTR:
-					dir = HTML_DIRECTION_LTR;
-					break;
-				default:
-					break;
-				}
+				if (child == HTML_CLUE (o)->head && html_object_get_length (child) == 0 && o->prev)
+					return html_object_get_direction (o->prev);
+
+				dir = html_text_direction_pango_to_html (html_text_get_pango_direction (HTML_TEXT (child)));
 
 				break;
 			}
